@@ -31,231 +31,101 @@
 
 from __future__ import annotations
 
-from collections import namedtuple as namedtuple_t
-from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
+import dataclasses as dtcl
+from typing import Any, Callable
+from typing import NamedTuple as named_tuple_t
+from typing import Sequence
 
 import numpy as nmpy
-import scipy.interpolate as in_
-import skimage.measure as ms_
+import scipy.interpolate as spnt
+import skimage.measure as sims
 
-import skl_graph.brick.elm_id as id_
-import skl_graph.brick.hierarchy as hrcy
-import skl_graph.type.topology_map as bymp
-from skl_graph.brick.constants import UNTESTED_VALIDITY
+import skl_graph.type.topology_map as tgmp
+from skl_graph.standard.identifier import COORDINATE_SEPARATOR, EncodedNumber
 from skl_graph.type.node import branch_node_t, end_node_t, node_t
 
 
 array_t = nmpy.ndarray
-# ww_length=width-weighted length
-# sq_lengths=squared lengths; Interest: all integers
-edge_lengths_t = namedtuple_t("edge_lengths_t", "length ww_length lengths sq_lengths")
 
 
+class measures_t(named_tuple_t):
+    """
+    segment_: prefix for pixel-to-pixel segments
+    segment_sq_lengths: squared segment lengths; Interest: all integers
+    segment_based_area: width-weighted length
+    """
+
+    length: float
+    segment_lengths: array_t
+    segment_sq_lengths: array_t
+    mean_width: float
+    mean_based_area: float | None
+    segment_based_area: float | None
+
+
+@dtcl.dataclass(slots=True, repr=False, eq=False)
 class raw_edge_t:
-    #
-    __slots__ = (
-        "dim",
-        "sites",
-    )
-
-    dim: int
-    sites: Tuple[array_t, ...]
-
-    def __init__(self):
-        #
-        for slot in self.__class__.__slots__:
-            setattr(self, slot, None)
+    sites: tuple[array_t, ...]
 
     @classmethod
-    def NewWithSites(cls, sites: Tuple[array_t, ...]) -> raw_edge_t:
-        #
-        instance = cls()
+    def NewWithSites(cls, sites: tuple[array_t, ...], /) -> raw_edge_t:
+        """"""
+        return cls(sites=_ReOrderedSites(sites))
 
-        instance.dim = sites.__len__()
-        instance.sites = _ReOrderedSites(sites)
-
-        return instance
+    @property
+    def dim(self) -> int:
+        """"""
+        return self.sites.__len__()
 
     @property
     def n_sites(self) -> int:
-        #
+        """"""
         return self.sites[0].size
 
     def __str__(self) -> str:
         """"""
         return (
             f"{self.__class__.__name__}:\n"
-            f"    Sites[{self.dim}-D]={self.sites[0].size}"
+            f"    Sites[{self.dim}-D]: {self.sites[0].size}"
         )
 
 
+@dtcl.dataclass(slots=True, repr=False, eq=False)
 class edge_t(raw_edge_t):
-    #
-    __slots__ = (
-        "uid",
-        "lengths",
-        "widths",
-        "invalidities",
-        "_cache",
-    )
-
-    uid: str
-    lengths: edge_lengths_t
-    widths: array_t
-    invalidities: List[str]  # Use brick.constants.UNTESTED_VALIDITY as initial value
-    _cache: Dict[str, Any]
-
-    def __init__(self):
-        #
-        super().__init__()
-        self.invalidities = UNTESTED_VALIDITY
-        self._cache = {}
+    uid: str = None
+    widths: array_t = None
+    measures: measures_t = None
+    _cache: dict[str, Any] = dtcl.field(default_factory=dict)
 
     @classmethod
     def NewWithDetails(
         cls,
-        sites: Tuple[array_t, ...],
+        sites: tuple[array_t, ...],
         adjacent_node_uids: Sequence[str],
+        /,
+        *,
         width_map: array_t = None,
     ) -> edge_t:
-        #
+        """"""
         raw_edge = raw_edge_t.NewWithSites(sites)
-        ip_edge = _building_edge_t.NewFromRaw(raw_edge)
+        transient = _transient_t.NewFromRaw(raw_edge)
 
-        ip_edge.SetUID(adjacent_node_uids)
-        ip_edge.AddWidths(width_map)
-        ip_edge.SetLengths(ip_edge.widths)
+        transient.SetUID(adjacent_node_uids)
+        transient.SetWidths(width_map)
+        transient.ComputeMeasures()
 
-        return ip_edge.AsEdge()
+        return transient.AsEdge()
 
-    @property
-    def is_probably_valid(self) -> bool:
-        """Partial validity test.
-
-        Missing validity test: The end sites must be found in the adjacent nodes position/sites, which can only be
-        tested at the graph level. See HasValidEndSites.
-
-        Returns
-        -------
-        bool
-
-        """
-        output = True
-        self.invalidities = []
-
-        n_sites = self.sites[0].size
-        length = self.lengths.length
-        sq_lengths = self.lengths.sq_lengths
-        if length < n_sites - 1:
-            output = False
-            self.invalidities.append(
-                f"{length}: Computed length cannot be smaller "
-                f"than number of sites - 1={n_sites - 1}"
-            )
-        if nmpy.any(sq_lengths == 0):
-            output = False
-            self.invalidities.append("Has repeated sites")
-        if nmpy.any(sq_lengths > self.dim):
-            output = False
-            self.invalidities.append("Has site gaps")
-
-        return output
-
-    def HasValidEndSites(
-        self,
-        origin: Union[end_node_t, branch_node_t],
-        destination: Union[end_node_t, branch_node_t],
-    ) -> bool:
-        """"""
-        end_sites = self.end_sites
-        n_found_end_sites = [0, 0]
-
-        for node in (origin, destination):
-            for e_idx in (0, 1):
-                if isinstance(node, end_node_t):
-                    end_site_found = nmpy.array_equal(end_sites[e_idx], node.position)
-                else:
-                    end_site = nmpy.reshape(end_sites[e_idx], (self.dim, 1))
-                    end_site_found = any(
-                        nmpy.all(end_site == nmpy.array(node.sites), axis=0)
-                    )
-                if end_site_found:
-                    n_found_end_sites[e_idx] += 1
-
-        if origin.uid == destination.uid:
-            n_expected = 2
-        else:
-            n_expected = 1
-
-        return n_found_end_sites == [n_expected, n_expected]
-
-    @property
-    def end_sites(self) -> Tuple[Tuple[Any, ...]]:
-        """"""
-        return tuple(
-            tuple(self.sites[idx_][_s_idx] for idx_ in range(self.dim))
-            for _s_idx in (0, -1)
-        )
-
-    def AsPolyline(self, max_distance: float = None) -> Tuple[float, array_t]:
-        """
-        Implementation of the Ramer–Douglas–Peucker algorithm
-        https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm
-        """
-        cache_entry = self.AsPolyline.__name__
-
-        if cache_entry not in self._cache:
-            self._cache[cache_entry] = {}
-
-        if max_distance is None:
-            if self._cache[cache_entry].__len__() > 0:
-                max_distance = min(self._cache[cache_entry].keys())
-            else:
-                max_distance = 1.0
-
-        if max_distance not in self._cache[cache_entry]:
-            output = nmpy.zeros(self.n_sites, dtype=nmpy.bool_)
-            output[0] = True
-            output[-1] = True
-
-            sites_as_array = nmpy.array(self.sites).T
-            segments = [(0, self.n_sites - 1)]
-            while segments.__len__() > 0:
-                first_idx, last_idx = segments.pop()
-                first_point = sites_as_array[first_idx, :]
-                along_line = sites_as_array[last_idx, :] - first_point
-                al_norm = nmpy.linalg.norm(along_line)  # al=along line
-
-                highest_distance = 0.0
-                idx_of_hd = first_idx  # hd=highest distance
-                for idx in range(first_idx + 1, last_idx):
-                    distance = _DistanceToLine(
-                        sites_as_array[idx, :], first_point, along_line, al_norm
-                    )
-                    if distance > highest_distance:
-                        idx_of_hd = idx
-                        highest_distance = distance
-
-                if highest_distance > max_distance:
-                    output[idx_of_hd] = True
-                    segments.extend(((first_idx, idx_of_hd), (idx_of_hd, last_idx)))
-
-            self._cache[cache_entry][max_distance] = (
-                max_distance,
-                nmpy.nonzero(output)[0],
-            )
-
-        return self._cache[cache_entry][max_distance]
-
-    def AsCurve(self) -> Optional[tuple]:
+    def AsCurve(self) -> tuple | None:
         """"""
         cache_entry = self.AsCurve.__name__
 
         if cache_entry not in self._cache:
             if self.n_sites > 1:
-                arc_lengths = nmpy.cumsum([0] + self.lengths.sq_lengths.tolist())
+                as_list: list[float] = list(self.measures.segment_lengths)
+                arc_lengths = nmpy.cumsum([0.0] + as_list)
                 self._cache[cache_entry] = tuple(
-                    in_.PchipInterpolator(arc_lengths, self.sites[idx_])
+                    spnt.PchipInterpolator(arc_lengths, self.sites[idx_])
                     for idx_ in range(self.dim)
                 )
             else:
@@ -263,15 +133,15 @@ class edge_t(raw_edge_t):
 
         return self._cache[cache_entry]
 
-    def OriginDirection(self) -> Optional[array_t]:
+    def OriginDirection(self) -> array_t | None:
         """"""
         return self._Direction(self.OriginDirection)
 
-    def FinalDirection(self) -> Optional[array_t]:
+    def FinalDirection(self) -> array_t | None:
         """"""
         return self._Direction(self.FinalDirection)
 
-    def _Direction(self, where: Callable) -> array_t:
+    def _Direction(self, where: Callable, /) -> array_t:
         """"""
         where = where.__name__
         if where not in self._cache:
@@ -280,7 +150,7 @@ class edge_t(raw_edge_t):
         return self._cache[where]
 
     def _SetEndPointDirections(self) -> None:
-        #
+        """"""
         cache_entry_o = self.OriginDirection.__name__
         cache_entry_f = self.FinalDirection.__name__
 
@@ -289,10 +159,10 @@ class edge_t(raw_edge_t):
             self._cache[cache_entry_o] = None
             self._cache[cache_entry_f] = None
         else:
-            max_arclength = as_curve[0].x.item(-1)
+            max_arc_length = as_curve[0].x.item(-1)
             o_dir, f_dir = [], []
             for d_idx in range(self.dim):
-                directions = as_curve[d_idx]((0, max_arclength), 1)
+                directions = as_curve[d_idx]((0, max_arc_length), 1)
                 o_dir.append(directions[0])
                 f_dir.append(directions[1])
             self._cache[cache_entry_o] = nmpy.array(o_dir, dtype=nmpy.float64) / (
@@ -302,48 +172,175 @@ class edge_t(raw_edge_t):
                 f_dir, dtype=nmpy.float64
             ) / nmpy.linalg.norm(f_dir)
 
-    def __str__(self) -> str:
+    def Curvatures(self) -> tuple[float, ...]:
         """"""
+        if self.dim != 2:
+            raise ValueError(
+                f"{self.dim}: Invalid dimension for curvature computation. Expected=2."
+            )
+
+        cache_entry = self.Curvatures.__name__
+
+        if cache_entry not in self._cache:
+            curvatures = []
+            self._cache[cache_entry] = tuple(curvatures)
+
+        return self._cache[cache_entry]
+
+    def CurvaturesAndTorsions(self) -> tuple[tuple[float, ...], tuple[float, ...]]:
+        """"""
+        if self.dim != 3:
+            raise ValueError(
+                f"{self.dim}: Invalid dimension for curvature and torsion computation. Expected=3."
+            )
+
+        cache_entry = self.CurvaturesAndTorsions.__name__
+
+        if cache_entry not in self._cache:
+            as_curve = self.AsCurve()
+            arc_lengths = as_curve[0].x
+            # derivatives_?: first index=0, 1, 2 for derivative orders 1, 2, 3; second index=arc-length parameter
+            derivatives_x, derivatives_y, derivatives_z = (
+                tuple(as_curve[_idx](arc_lengths, _rdr) for _rdr in (1, 2, 3))
+                for _idx in range(self.dim)
+            )
+            # key=space dimension index among 1, 2, 3 (i.e., X, Y, Z)
+            # value=list of first+second+third derivatives for each curve parameter
+            # derivatives: dict[int, list[tuple[float, float, float]]] = {}
+            # for d_idx in range(self.dim):
+            #     per_dim = []
+            #     for parameter in arc_lengths:
+            #         derivative_1 = as_curve[d_idx](parameter, 1).item()
+            #         derivative_2 = as_curve[d_idx](parameter, 2).item()
+            #         derivative_3 = as_curve[d_idx](parameter, 3).item()
+            #         per_dim.append((derivative_1, derivative_2, derivative_3))
+            #     derivatives[d_idx + 1] = per_dim
+
+            a = (
+                derivatives_z[1] * derivatives_y[0]
+                - derivatives_y[1] * derivatives_z[0]
+            )
+            b = (
+                derivatives_x[1] * derivatives_z[0]
+                - derivatives_z[1] * derivatives_x[0]
+            )
+            c = (
+                derivatives_y[1] * derivatives_x[0]
+                - derivatives_x[1] * derivatives_y[0]
+            )
+            abc_sq_length = a**2 + b**2 + c**2
+
+            normalization = (
+                derivatives_x[0] ** 2 + derivatives_y[0] ** 2 + derivatives_z[0] ** 2
+            ) ** 1.5
+            where_safe = normalization != 0.0
+            curvatures = nmpy.empty_like(arc_lengths)
+            curvatures[where_safe] = (
+                nmpy.sqrt(abc_sq_length[where_safe]) / normalization[where_safe]
+            )
+            curvatures[nmpy.logical_not(where_safe)] = 0.0
+
+            where_safe = abc_sq_length != 0.0
+            torsions = nmpy.empty_like(arc_lengths)
+            torsions[where_safe] = nmpy.abs(
+                (
+                    derivatives_x[2][where_safe] * a[where_safe]
+                    + derivatives_y[2][where_safe] * b[where_safe]
+                    + derivatives_z[2][where_safe] * c[where_safe]
+                )
+                / abc_sq_length[where_safe]
+            )
+            torsions[nmpy.logical_not(where_safe)] = 0.0
+
+            # curvatures = []
+            # torsions = []
+            # for p_idx in range(arc_lengths.__len__()):
+            #     a = (
+            #         derivatives_z[1][p_idx] * derivatives_y[0][p_idx]
+            #         - derivatives_y[1][p_idx] * derivatives_z[0][p_idx]
+            #     )
+            #     b = (
+            #         derivatives_x[1][p_idx] * derivatives_z[0][p_idx]
+            #         - derivatives_z[1][p_idx] * derivatives_x[0][p_idx]
+            #     )
+            #     c = (
+            #         derivatives_y[1][p_idx] * derivatives_x[0][p_idx]
+            #         - derivatives_x[1][p_idx] * derivatives_y[0][p_idx]
+            #     )
+            #     abc_sq_length = a**2 + b**2 + c**2
+            #
+            #     normalization = (
+            #         derivatives_x[0][p_idx] ** 2
+            #         + derivatives_y[0][p_idx] ** 2
+            #         + derivatives_z[0][p_idx] ** 2
+            #     ) ** 1.5
+            #     if normalization != 0.0:
+            #         curvature = (nmpy.sqrt(abc_sq_length) / normalization).item()
+            #     else:
+            #         curvature = 0.0
+            #
+            #     if abc_sq_length != 0.0:
+            #         torsion = nmpy.abs(
+            #             (
+            #                 derivatives_x[2][p_idx] * a
+            #                 + derivatives_y[2][p_idx] * b
+            #                 + derivatives_z[2][p_idx] * c
+            #             )
+            #             / abc_sq_length
+            #         ).item()
+            #     else:
+            #         torsion = 0.0
+            #
+            #     curvatures.append(curvature)
+            #     torsions.append(torsion)
+
+            self._cache[cache_entry] = (tuple(curvatures), tuple(torsions))
+
+        return self._cache[cache_entry]
+
+    def __str__(self) -> str:
+        """
+        Does not call raw_edge_t.__str__ since raw edges do not have UID
+        """
+        origin = tuple(self.sites[_idx][0] for _idx in range(self.dim))
+        if self.measures is None:
+            raw_length = "Not computed yet"
+            segment_based_area = raw_length
+        else:
+            raw_length = round(self.measures.length, 2)
+            segment_based_area = round(self.measures.segment_based_area, 2)
+
+        main = (
+            f"{self.__class__.__name__}.{self.uid}:\n"
+            f"    Sites[{self.dim}-D]: {self.sites[0].size}\n"
+            f"    Origin:              {origin}\n"
+            f"    Lengths:             Raw={raw_length}, SB-Area={segment_based_area}"
+        )
+
         if (self._cache is None) or (self._cache.__len__() == 0):
             cached_values = "None yet"
         else:
             cached_values = ", ".join(self._cache.keys())
 
-        return _EdgeCommonDescription(self) + f"\n    Cached values: {cached_values}"
+        return main + f"\n    Cached values: {cached_values}"
 
 
-class _building_edge_t(raw_edge_t):
-    #
-    __slots__ = (
-        "uid",
-        "lengths",
-        "widths",
-    )
-
-    uid: str
-    lengths: edge_lengths_t
-    widths: array_t
+@dtcl.dataclass(slots=True, repr=False, eq=False)
+class _transient_t(raw_edge_t):
+    uid: str = None
+    widths: array_t = None
+    measures: measures_t = None
 
     @classmethod
-    def NewFromRaw(cls, raw_edge: raw_edge_t) -> _building_edge_t:
-        #
-        instance = cls()
-
-        for slot in hrcy.AllSlotsOfClass(raw_edge.__class__):
-            setattr(instance, slot, getattr(raw_edge, slot))
-
-        return instance
+    def NewFromRaw(cls, raw_edge: raw_edge_t, /) -> _transient_t:
+        """"""
+        return cls(**dtcl.asdict(raw_edge))
 
     def AsEdge(self) -> edge_t:
-        #
-        output = edge_t()
+        """"""
+        return edge_t(**dtcl.asdict(self))
 
-        for slot in hrcy.AllSlotsOfClass(self.__class__):
-            setattr(output, slot, getattr(self, slot))
-
-        return output
-
-    def SetUID(self, adjacent_node_uids: Sequence[str]) -> None:
+    def SetUID(self, adjacent_node_uids: Sequence[str], /) -> None:
         """"""
         if adjacent_node_uids.__len__() != 2:
             raise RuntimeError(
@@ -355,50 +352,60 @@ class _building_edge_t(raw_edge_t):
             node_uid_0, node_uid_1 = node_uid_1, node_uid_0
 
         uid_components = [
-            id_.EncodedNumber(coord)
-            for coord in node_uid_0.split(id_.COORDINATE_SEPARATOR)
+            EncodedNumber(coord) for coord in node_uid_0.split(COORDINATE_SEPARATOR)
         ]
-        uid_components.append(id_.COORDINATE_SEPARATOR)
+        uid_components.append(COORDINATE_SEPARATOR)
         uid_components.extend(
-            id_.EncodedNumber(coord)
-            for coord in node_uid_1.split(id_.COORDINATE_SEPARATOR)
+            EncodedNumber(coord) for coord in node_uid_1.split(COORDINATE_SEPARATOR)
         )
 
         self.uid = "".join(uid_components)
 
-    def SetLengths(self, widths: array_t) -> None:
+    def SetWidths(self, width_map: array_t, /) -> None:
+        """"""
+        if width_map is not None:
+            self.widths = width_map[self.sites]
+
+    def ComputeMeasures(self) -> None:
         """
-        Passing widths instead of using self.widths ensures that this method will not be called before setting
-        self.widths, should it be set.
+        Must not be called before SetWidths
         """
         sites_as_array = nmpy.array(self.sites)
         segments = nmpy.diff(sites_as_array, axis=1)
-        sq_lengths = (segments**2).sum(axis=0)
-        lengths = nmpy.sqrt(sq_lengths)
-        length = lengths.sum().item()
+        segment_sq_lengths = nmpy.sum(segments**2, axis=0)
+        segment_lengths = nmpy.sqrt(segment_sq_lengths)
+        length = segment_lengths.sum().item()
 
-        if widths is None:
-            ww_length = -1.0
+        if self.widths is None:
+            mean_width = mean_based_area = segment_based_area = 0.0
         else:
-            ww_length = (0.5 * (widths[1:] + widths[:-1]) * lengths).sum().item()
+            mean_width = nmpy.mean(self.widths).item()
+            mean_based_area = length * mean_width
+            segment_based_area = (
+                (0.5 * (self.widths[1:] + self.widths[:-1]) * segment_lengths)
+                .sum()
+                .item()
+            )
 
-        self.lengths = edge_lengths_t(
-            length=length, ww_length=ww_length, lengths=lengths, sq_lengths=sq_lengths
+        self.measures = measures_t(
+            length=length,
+            segment_lengths=segment_lengths,
+            segment_sq_lengths=segment_sq_lengths,
+            mean_width=mean_width,
+            mean_based_area=mean_based_area,
+            segment_based_area=segment_based_area,
         )
 
-    def AddWidths(self, width_map: array_t) -> None:
-        #
-        if width_map is not None:
-            self.widths = width_map[self.sites]
-
     def AppendBranchNode(
         self,
         b_coords: array_t,
         node: node_t,
-        adjacent_node_uids: List[str],
+        adjacent_node_uids: list[str],
+        /,
+        *,
         force_after: bool = False,
     ) -> None:
-        #
+        """"""
         adjacent_node_uids.append(node.uid)
 
         space_dim = self.dim
@@ -428,40 +435,18 @@ class _building_edge_t(raw_edge_t):
                 for idx_ in range(space_dim)
             )
 
-    def __str__(self) -> str:
-        """"""
-        return _EdgeCommonDescription(self)
-
-
-def _EdgeCommonDescription(edge: Union[edge_t, _building_edge_t]) -> str:
-    """"""
-    origin = tuple(edge.sites[idx][0] for idx in range(edge.dim))
-    if edge.lengths is None:
-        raw_length = "Not computed yet"
-        ww_length = raw_length
-    else:
-        raw_length = round(edge.lengths.length, 2)
-        ww_length = round(edge.lengths.ww_length, 2)
-
-    return (
-        f"{edge.__class__.__name__}[{edge.uid}]:\n"
-        f"    Sites[{edge.dim}-D]={edge.sites[0].size}\n"
-        f"    Origin: {origin}\n"
-        f"    Lengths: Raw={raw_length}, WW={ww_length}"
-    )
-
 
 def RawEdges(
-    skl_map: array_t, b_node_lmap: array_t
-) -> Tuple[Sequence[raw_edge_t], array_t]:
+    skl_map: array_t, b_node_lmap: array_t, /
+) -> tuple[Sequence[raw_edge_t], array_t]:
     """"""
     edge_map = skl_map.astype(nmpy.int8)
     edge_map[b_node_lmap > 0] = 0
-    edge_lmap, n_edges = bymp.LABELING_FCT_FOR_DIM[skl_map.ndim](edge_map)
+    edge_lmap, n_edges = tgmp.LABELING_FCT_FOR_DIM[skl_map.ndim](edge_map)
 
-    edge_props = ms_.regionprops(edge_lmap)
+    edge_props = sims.regionprops(edge_lmap)
 
-    edges = n_edges * [raw_edge_t()]
+    edges: list[raw_edge_t] = n_edges * [None]
     for props in edge_props:
         sites = props.image.nonzero()
         for d_idx in range(skl_map.ndim):
@@ -478,32 +463,33 @@ def EdgesFromRawEdges(
     edge_lmap: array_t,
     e_node_lmap: array_t,
     b_node_lmap: array_t,
+    /,
+    *,
     width_map: array_t = None,
-) -> Tuple[Tuple[edge_t], List[List[str]]]:
-    #
-    edge_tmap = bymp.TopologyMapOfMap(edge_lmap > 0)
-    # ip_=in progress
-    ip_edges = [_building_edge_t.NewFromRaw(edge) for edge in raw_edges]
+) -> tuple[tuple[edge_t], list[list[str]]]:
+    """"""
+    edge_tmap = tgmp.TopologyMapOfMap(edge_lmap > 0)
+    transient_edges = [_transient_t.NewFromRaw(_raw) for _raw in raw_edges]
 
     # ep=edge end point; Keep < 2 since ==0 (length-1 edges) and ==1 (other edges) are needed
     # Do not use list multiplication since the same list then used for all the elements
-    node_uids_per_edge: list[list[str]] = [[] for _ in ip_edges]
+    node_uids_per_edge: list[list[str]] = [[] for _ in transient_edges]
     for ep_coords in zip(*(edge_tmap < 2).nonzero()):
         edge_idx = edge_lmap[ep_coords] - 1
-        edge = ip_edges[edge_idx]
+        transient = transient_edges[edge_idx]
         e_node_label = e_node_lmap[ep_coords]
 
         if e_node_label > 0:
             # End node-to-X edge (i.e., edge end point is also an end node)
             node_uids_per_edge[edge_idx].append(e_nodes[e_node_label - 1].uid)
-            if edge.n_sites == 1:
+            if transient.n_sites == 1:
                 # End node-to-branch node edge (and there is a unique non-zero value in b_neighborhood)
                 nh_slices_starts, b_neighborhood = _LMapNeighborhood(
                     b_node_lmap, ep_coords
                 )
                 b_node_label = nmpy.amax(b_neighborhood)
                 b_coords = nmpy.transpose((b_neighborhood == b_node_label).nonzero())[0]
-                edge.AppendBranchNode(
+                transient.AppendBranchNode(
                     nmpy.add(nh_slices_starts, b_coords),
                     b_nodes[b_node_label - 1],
                     node_uids_per_edge[edge_idx],
@@ -514,7 +500,7 @@ def EdgesFromRawEdges(
             # Looping only for length-1, b-to-b edges
             for b_coords in zip(*b_neighborhood.nonzero()):
                 b_node_label = b_neighborhood[b_coords]
-                edge.AppendBranchNode(
+                transient.AppendBranchNode(
                     nmpy.add(nh_slices_starts, b_coords),
                     b_nodes[b_node_label - 1],
                     node_uids_per_edge[edge_idx],
@@ -522,27 +508,43 @@ def EdgesFromRawEdges(
                 )
                 force_after = not force_after
 
-    for edge, adjacent_node_uids in zip(ip_edges, node_uids_per_edge):
-        edge.SetUID(adjacent_node_uids)
-        edge.AddWidths(width_map)
-        edge.SetLengths(edge.widths)
+    for transient, adjacent_node_uids in zip(transient_edges, node_uids_per_edge):
+        # Fix sites order for self-loops connected to branch nodes
+        if adjacent_node_uids[0] == adjacent_node_uids[1]:
+            node = None
+            for current in b_nodes:
+                if current.uid == adjacent_node_uids[0]:
+                    node = current
+                    break
+            if node is not None:
+                node_sites = nmpy.array(node.sites)
+                edge_sites = nmpy.array(transient.sites)
+                while not all(
+                    any(
+                        nmpy.all(
+                            (node_sites - edge_sites[:, _sdx][:, None]) == 0, axis=0
+                        )
+                    )
+                    for _sdx in (0, -1)
+                ):
+                    edge_sites = nmpy.roll(edge_sites, -1, axis=1)
 
-    edges = tuple(edge.AsEdge() for edge in ip_edges)
+                transient.sites = tuple(
+                    edge_sites[_cdx, :] for _cdx in range(transient.dim)
+                )
+
+        transient.SetUID(adjacent_node_uids)
+        transient.SetWidths(width_map)
+        transient.ComputeMeasures()
+
+    edges = tuple(_tst.AsEdge() for _tst in transient_edges)
 
     return edges, node_uids_per_edge
 
 
-def _ReOrderedSites(sites: Tuple[array_t, ...]) -> Tuple[array_t, ...]:
+def _ReOrderedSites(sites: tuple[array_t, ...], /) -> tuple[array_t, ...]:
     """
     If the number of sites is 1 or 2, the input argument is returned (i.e., no copy is made).
-
-    Parameters
-    ----------
-    sites
-
-    Returns
-    -------
-
     """
     n_sites = sites[0].size
     if n_sites < 3:
@@ -624,12 +626,7 @@ def _ReOrderedSites(sites: Tuple[array_t, ...]) -> Tuple[array_t, ...]:
 
     reordered_coords = nmpy.array(reordered_coords)
     if self_loop:
-        self_origin_idx = nmpy.argwhere(
-            nmpy.all(reordered_coords == self_origin, axis=1)
-        ).item()
-        if self_origin_idx > 0:
-            reordered_coords = nmpy.roll(reordered_coords, -self_origin_idx, axis=0)
-        reordered_coords = nmpy.vstack((reordered_coords, self_origin))
+        reordered_coords = _RolledSitesWithFixedOrigin(reordered_coords, self_origin)
     reordered_coords = tuple(reordered_coords[:, _idx] for _idx in range(dim))
 
     return reordered_coords
@@ -637,7 +634,7 @@ def _ReOrderedSites(sites: Tuple[array_t, ...]) -> Tuple[array_t, ...]:
 
 def _NeighborIndices(
     dim: int, sites: array_t, unvisited_slc: array_t, end_point: array_t
-) -> Tuple[array_t, array_t]:
+) -> tuple[array_t, array_t]:
     """"""
     unvisited_sites = sites[unvisited_slc, :]
 
@@ -647,18 +644,17 @@ def _NeighborIndices(
     return neighbor_idc, unvisited_sites
 
 
-def _DistanceToLine(
-    point: array_t, on_line: array_t, along_line: array_t, al_norm: float
-) -> float:
-    #
-    if al_norm == 0.0:
-        return nmpy.linalg.norm(point - on_line)
+def _RolledSitesWithFixedOrigin(sites: array_t, origin: array_t, /) -> array_t:
+    """"""
+    self_origin_idx = nmpy.argwhere(nmpy.all(sites == origin, axis=1)).item()
+    if self_origin_idx > 0:
+        sites = nmpy.roll(sites, -self_origin_idx, axis=0)
 
-    return nmpy.linalg.norm(nmpy.cross(point - on_line, along_line)) / al_norm
+    return nmpy.vstack((sites, origin))
 
 
-def _LMapNeighborhood(lmap: array_t, site: Tuple[int, ...]) -> Tuple[array_t, array_t]:
-    #
+def _LMapNeighborhood(lmap: array_t, site: tuple[int, ...]) -> tuple[array_t, array_t]:
+    """"""
     slices_starts = tuple(max(site[idx_] - 1, 0) for idx_ in range(site.__len__()))
     slices = tuple(
         slice(slices_starts[idx_], min(site[idx_] + 2, lmap.shape[idx_]))
